#!/bin/bash -u

# ----------------------------------------------------------------
# File Name:           CFHTLS_datarequest.sh
# Author:              Thomas Erben (terben@astro.uni-bonn.de)
# Last modified on:    17.02.2010
# Description:         Requests and retrievs public CFHT data 
#                      from CADC
# Version:             $$VERSION$$
# ----------------------------------------------------------------

# Script history information:
#
# - 03.02.2007:
#   I changed the query whether a database file was given.
#   The condition '-n ${}' seems to be valid for an empty
#   (being "") string.
#
# - 29.06.2007:
#   New public Elixir images (requested after 29.06.2007)
#   seem to have only four error/warning messages with
#   'fitsverify'. I adapted the image checking accordingly.
#
# - 04.07.2007:
#   Files that need to be retrieved but are already present
#   in the output directory are deleted first! Those are
#   files whose previous transfer probably failed and hence
#   the file did not make it to the database file.
#
# - 06.07.2007:
#   I encountered files who have in all extensions a 
#   DATEEND= 'NULL' keyword whose value is reported as 
#   illegal by fitsverify. I modified the error checking
#   such that these files pass the fitsverify test and
#   are added to the database file.
#
# - 21.07.2007:
#   I corrected an error when the script is called with
#   a databsse file which does not yet exist. In this case
#   all images meeting the initial criteria are listed/requested.
#
# - 28.07.2007:
#   Data are now requested with a new address which needs
#   authorication with CADC USER/PASSWORD. It seems that the
#   old anonymous address does not work anymore. Hence new
#   command line options for HTTP User and Password.
#
# - 02.10.2007:
#   Update of documentation.
#
# - 29.10.2007:
# - I introduced the possibility ro retrieve data via an image
#   list.
# - If no user/passwd combination is given we try to retrieve via
#   an anonymous server.
#
# - 23.11.2007:
#   I corrected a bug in the command line options; I used '-u' to set
#   the CADC password instead of '-p'.
#
# - 13.12.2007:
#   I encountered CFHTLS FITS files with 23 fitsverify warnings (depreceated
#   EPOCH keywords only in some extensions but not in all). I modified
#   the acceptance condition for FITS files accordingly.
#
# - 24.04.2008:
#   I introduced a check whether retrieved images have 36 extensions before
#   adding them to the database.
#
# - 27.04.2008:
#   I only issue a warning if the wget call exits with an error status;
#   before it was an error with exiting the script. It can happen that
#   some files are successfully retieved before wget exits. These files
#   then would not be checked and added to a database file.
#
# - 17.04.2009:
#   I extended the script so that it can be used for other CFHT instruments
#   except MEGAPRIME; currently WIRCAM and CFH12K. The instrument
#   (if it is not MEGAPRIME) needs to be manually specified via the 
#   '-in' command line option (SUBOPTIMAL; better would be an automatic
#   detection).
#
# - 28.07.2009:
#   documentation update
#
# - 05.01.2010:
#   script extension to retrieve NGVS data with a THELI hand-made NGVS
#   quality catalogue
#
# - 10.01.2010:
#   - It is now mandatory to supply the instrument on the command line!
#     We can no longer forget to set it!
#   - If images are Rice-compressed (instead gzip), the output from 
#     'fitsverify' differs. This needs to be taken into account in 
#     the image checking phase.
#
# - 15.02.2010:
#   Update of the script to the 'official' CADC retrieval address
#   (see http://www3.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/getData/doc)
#   for data queries. Thios brings a speed-up of abour a factor of 2!
#
# - 17.02.2010:
#   get rid of empty result files in case of 'wget' failures

# TODO:
# Automatic detection of instrument!

# preliminary work

# define THELI_DEBUG because of the '-u' script flag
# (the use of undefined variables will be treated as errors!)
THELI_DEBUG=${THELI_DEBUG:-""}

# function definitions
function printUsage 
{
  echo "SCRIPT NAME:"
  echo "    CFHTLS_datarequest.sh - retrieve CFHT data from CADC"
  echo ""
  echo "SYNOPSIS:"
  echo "    CFHTLS_datarequest.sh -in CFHT_instrument (MEGAPRIME, CFH12K, WIRCAM)"
  echo "                          [ -fi CFHTLS_Field (D1-D4 and W1-W3) ]"
  echo "                          [ -f Filter (u,g,r,i,z) ]"
  echo "                          [ -i image list ]"
  echo "                          [ -u CADC_User ]"
  echo "                          [ -p CADC_Password ]"
  echo "                          [ -e minimum_exposure_time (s) ]"
  echo "                          [ -ts Time_Start (dd/mm/yyyy) ]"
  echo "                          [ -te Time_End (dd/mm/yyyy) ]"
  echo "                          [ -l Only list files but do not retrieve ]"
  echo "                          [ -o Output_Directory ]"
  echo "                          [ -c Image_Catalogue ]"
  echo "                          [ -d Database_File ]"
  echo ""
  echo "    '-in' is the only mandatory option."
  echo ""
  echo "IMPORTANT SECURITY NOTE (READ CAREFULLY !!!!): "
  echo "    This script requires your CADC user/password combination on the"
  echo "    command line if you need to retrieve data from a non-anonymous server. "
  echo "    While the script is running these data appear in"
  echo "    'CLEAR TEXT' in the process list of your machine. Hence, other"
  echo "    users can get this information with simple process tracing"
  echo "    utilities such as 'ps'."
  echo ""
  echo "    If you cannot live with this DO NOT USE this script in its current"
  echo "    form."
  echo ""
  echo "    One possible solution is to hardcode your user/passwd combination"
  echo "    in the corresponding 'wget' call in this file and not to provide"
  echo "    them on the command line. Afterwards, give the script read-,"
  echo "    write- and execute permission for yourself only.  In contrast to"
  echo "    this script, the 'wget' call itself does not provide sensible data on"
  echo "    the process list."
  echo ""
  echo "DESCRIPTION: "
  echo "    The script can automatically retrieve public data from the CFHT"
  echo "    archive at CADC. It can be used to get all images from a specified"
  echo "    list ('-i' option). In this case each line of the input list has"
  echo "    to contain the name of one image (see below for an example).  If"
  echo "    '-i' is given, all options except '-in', '-u', '-p' and '-d' have no"
  echo "    effect. The CFHT instrument from which data should be requested"
  echo "    needs to be specified with the '-in' option; we currently support"
  echo "    MEGAPRIME, WIRCAM and CFH12K. If a database file of already"
  echo "    existing files is present (option '-d'; this is a simple ASCII"
  echo "    file containing an image name at each line) the images to be"
  echo "    requested are checked against this list and double images are"
  echo "    rejected from the request process. Images are then automatically"
  echo "    retrieved from CADC via wget and put in the current directory."
  echo "    Retrieved FITS files are checked for integrity with the "
  echo "    'fitsverify' tool."
  echo "    The User/Password combination of a valid CADC User has to be "
  echo "    provided if you need to retrieve proprietary data. For public"
  echo "    data, no authentification is necessary!"
  echo "    "
  echo "    The original main intention of the script is to 'automatically'"
  echo "    retrieve public data from the CFHTLS and NGVS surveys (raw Elixir"
  echo "    preprocessed images from the MEGAPRIME@CFHT instrument) from"
  echo "    CADC. It uses an image catalogue as created by"
  echo "    'CFHTLS_quality_catalog.sh' or 'do_NGVS_cat.sh'. The user provides"
  echo "    a CFHTLS/NGVS field (D1-D4, W1-W4, NGVS; also indiviudal Wide"
  echo "    fields such as W1p2p2 or NGVSm0m4 can be given), filter (ugriyz),"
  echo "    start and end observing date and minimum exposure time from the"
  echo "    images that should be requested. The imagenames are then extracted"
  echo "    from the image catalogue and, if present, checked against a"
  echo "    database file.  Requested data are put to a directory with the"
  echo "    name 'Output_directory/field/filter' and verified for integrity"
  echo "    with 'fitsverify'."
  echo ""
  echo "    Default values for not passed command line arguments are:"
  echo "    CFHTLS_field:            D1"
  echo "    Filter:                  i"
  echo "    minimum_exposure_time:   100s"
  echo "    Output_Directory:        current directory"
  echo "    Time_Start:              01/03/2003"
  echo "    Time_End:                31/12/2010"
  echo "    Image_Catalogue          ./Deep_Wide.cat"
  echo ""
  echo "    In all circumstances, the script needs the UNIX tools"
  echo "    'awk', 'sort', 'uniq', 'cat' and 'wget'. It also requires  the "
  echo "    'fitsverify' tool in your path."
  echo "    (see http://heasarc.gsfc.nasa.gov/docs/software/ftools/fitsverify/) "
  echo ""
  echo "    If you intend to use the script for CFHTLS/NGVS data retrieval you need,"
  echo "    in addition, the THELI pipeline installed "
  echo "    (see http://www.astro.uni-bonn.de/~mischa/theli.html), and an CFHTLS/NGVS "
  echo "    image catalogue (you can send a request to the script author to obtain"
  echo "    those)."
  echo " "
  echo "EXAMPLES:"
  echo "    - The file 'images.list' contains the two lines:"
  echo "      "
  echo "      816620p"
  echo "      816772p"
  echo ""
  echo "      The data originate from the MEGAPRIME instrument."
  echo "      The call 'CFHTLS_datarequest.sh -in MEGAPRIME -i images.list -d data.txt'"
  echo "      would first check whether these two files are listed in the"
  echo "      database file 'data.txt'. If not, they are retrieved from CADC"
  echo "      and checked for integrity. If the images pass this test they are"
  echo "      appended to 'data.txt'."
  echo ""
  echo "    - CFHTLS_datarequest.sh -in MEGAPRIME -ts 30/05/2004 \\"
  echo "                            -te 29/05/2005 -fi W1 -f z"
  echo ""
  echo "      All z-band images located in the CFHTLS W1 area and "
  echo "      observed between 30/05/2004 and 29/05/2005 are retrieved from CADC."
  echo ""
  echo "KNOWN PROBLEMS:"
  echo "    The script does not check whether the requested data originate from the"
  echo "    specified instrument. If there is a mismatch the script will give"
  echo "    unpredictable results in the file verification phase."
  echo ""
  echo "AUTHOR:"
  echo "    Thomas Erben (terben@astro.uni-bonn.de)"
  echo ""
}

function cleanTmpFiles 
{
  if [ -z ${THELI_DEBUG} ]; then
    echo "Cleaning temporary files for script $0"
    test -f reqfiles_$$.txt          && rm reqfiles_$$.txt
    test -f cataloguefiles_$$.txt    && rm cataloguefiles_$$.txt
    test -f wget_$$.txt              && rm wget_$$.txt
    test -f data_uniq_$$.txt         && rm data_uniq_$$.txt
    test -f catalogue_uniq_$$.txt    && rm catalogue_uniq_$$.txt
    test -f dupl_images_$$.txt       && rm dupl_images_$$.txt
  else
    echo "Variable THELI_DEBUG set! No cleaning of temp. files in script $0"
  fi
}

# Handling of program interruption by CRTL-C
trap "echo 'Script $0 interrupted!! Cleaning up and exiting!'; \
      cleanTmpFiles; exit 0" INT

# Default values for command line arguments
IMAGE_CATALOGUE="./Deep_Wide.cat"
OUTPUT_DIR=`pwd`
DATABASE_FILE=""
MIN_EXPTIME=100
FILTER="i"
FIELD="D1"
INSTRUMENT="DUMMY"
TIME_START=01/03/2003
TIME_END=31/12/2010
LIST_ONLY=N
CADC_USER="DUMMY"
CADC_PASSWORD="DUMMY"
INPUT_IMAGE_LIST="DUMMY"
TRY_ANONYMOUS=0

if [ $# -lt 1 ]
then
  printUsage
  exit 0
fi

# Parse command line arguments
while [ $# -gt 0 ]
do
  case $1 in
  -c)
    IMAGE_CATALOGUE=${2}
    shift 2
    ;;
  -d)
    DATABASE_FILE=${2}
    shift 2
    ;;
  -e)       
    MIN_EXPTIME=${2}
    shift 2
    ;; 
  -f)
    FILTER=${2}
    shift 2
    ;;
  -fi)
    FIELD=${2}
    shift 2
    ;; 
  -h)
    printUsage
    exit 0
    ;;
  -i)
    INPUT_IMAGE_LIST=${2}
    shift 2
    ;;
  -in)
    INSTRUMENT=${2}
    shift 2
    ;;
  -l)
    LIST_ONLY=Y
    shift
    ;;
  -o)
    OUTPUT_DIR=${2}
    shift 2
    ;;
  -ts)
    TIME_START=${2}
    shift 2
    ;;
  -te)
    TIME_END=${2}
    shift 2
    ;;
  -u)
    CADC_USER=${2}
    shift 2
    ;;
  -p)
    CADC_PASSWORD=${2}
    shift 2
    ;;
   *)
    printUsage
    exit 1	   
    ;;
  esac
done

# check validity of instrument:
if [ "${INSTRUMENT}" = "DUMMY" ]
then
  echo "You need to specify the CFHT instrument!"
  echo "Currently I support MEGAPRIME, CFH12K and WIRCAM"
  exit 1;
fi

if [ "${INSTRUMENT}" != "MEGAPRIME" ] && [ "${INSTRUMENT}" != "WIRCAM" ] && \
   [ "${INSTRUMENT}" != "CFH12K" ]
then
  echo "Instrument ${INSTRUMENT} is unknown to me! Exiting!"
  exit 1;
fi 

# The expected number of image extensions and
# error codes for an instrument:
if [ "${INSTRUMENT}" = "MEGAPRIME" ]; then
  # The number of FITS extensions (chips) of the instrument:  
  INST_EXTENSIONS=36

  # The MEGAPRIME files we investigated so far produce either 41
  # `fitsverify` warnings (depreceated EQUINOX keywords in all
  # extensions and CTYPE and CRVAL keywords in the primary HDU) or
  # only four.  In the latter case the EQUINOX keys in the extensions
  # are not present.  Therefore we mark images producing 0, 4 or 41 as
  # exit code (`fitsverify` gives the total number of warnings and
  # errors as exit code) as ok.  Similar 'errors' (23, 78) with
  # o.k. images occured in the meanwhile:
  ERROR_CODES='"0 4 23 41 78"'
fi
if [ "${INSTRUMENT}" = "WIRCAM" ]; then
  INST_EXTENSIONS=4
  ERROR_CODES='"0 4 8"'
fi
if [ "${INSTRUMENT}" = "CFH12K" ]; then
  INST_EXTENSIONS=12
  ERROR_CODES='"0 1 37"'
fi

# check that User/Password has been provided:
if [ "${CADC_USER}" = "DUMMY" ] || [ "${CADC_PASSWORD}" = "DUMMY" ]
then
  echo "A CADC User and Password has not been provided !!"
  echo "I will try to retrieve data anonymously!"
  TRY_ANONYMOUS=1
fi 

if [ "${INPUT_IMAGE_LIST}" != "DUMMY" ]; then
  test -f ${INPUT_IMAGE_LIST}\
    || { echo "List ${INPUT_IMAGE_LIST} not present !!"; exit 1; }  

  sort ${INPUT_IMAGE_LIST} | uniq > cataloguefiles_$$.txt
else 
  # convert dates to GaBoDS IDs:
  MJD=`mjd -t 23:59:59 -d ${TIME_START} | awk '{print $7}'`
  GABODSID_START=`nightid -t 22:00:00 -d 31/12/1998 -m ${MJD} |\
                  awk 'NR > 1 {print int($6)}'`
  MJD=`mjd -t 23:59:59 -d ${TIME_END} | awk '{print $7}'`
  GABODSID_END=`nightid -t 22:00:00 -d 31/12/1998 -m ${MJD} |\
                awk 'NR > 1 {print int($6)}'`
  
  # extract from the image catalogue all files that fulfill the selection 
  # criteria:
  test -f ${IMAGE_CATALOGUE} || \
       { echo "${IMAGE_CATALOGUE} does not exist!!"; exit 1; }

  # check whether we have a CFHTLS, or an NGVS catalogue:
  ldactestexist -i ${IMAGE_CATALOGUE} -t OBJECTS -k CFHTLSFIELD
  if [ $? -eq 0 ]; then
    CATFIELDS="IMAGENAME CFHTLSFIELD FILTER GABODSID EXPTIME"
  else
    ldactestexist -i ${IMAGE_CATALOGUE} -t OBJECTS -k NGVSFIELD
    if [ $? -eq 0 ]; then
      CATFIELDS="IMAGENAME NGVSFIELD FILTER GABODSID EXPTIME"
    else
      echo "Your catalogue ${IMAGE_CATALOGUE} contains neither"
      echo "of the keys CFHTLSFIELD or NGVSFIELD!"
      exit 1;
    fi 
  fi

  ldactoasc -i ${IMAGE_CATALOGUE} -t OBJECTS -s -k ${CATFIELDS} |\
      awk '$2 ~ /^'${FIELD}'/ && 
           $4 > '${GABODSID_START}' && 
           $4 < '${GABODSID_END}' && 
           $3 == "'${FILTER}'" && $5 > '${MIN_EXPTIME}' {print $1"p"}' >\
      cataloguefiles_$$.txt 
fi

# Are there any files meeting the criteria
test -s cataloguefiles_$$.txt ||\
    { echo "No files match your criteria!! Exiting!!"; cleanTmpFiles; exit 1; }

echo "$(wc -l cataloguefiles_$$.txt | awk '{print $1}') images\
 meet your initial selection criteria!"

# check against the database file whether we have some files
# already

if [ "${DATABASE_FILE}" != "" ]
then
  if [ -f ${DATABASE_FILE} ] 
  then
    # compare initial image list with database file and
    # reject images that are already present.
    # Tactic: concatenate the files (that are first sorted and
    # uniqified) to compare and identify double lines. 
    # Those double lines are removed from the files to
    # be requested. The initial uniqifying is just a
    # security measure if for some reason double entries
    # are present in image list or database.
    echo "Comparing initial image list with database file ${DATABASE_FILE}!"
    #
    # The sed commands deal with unwanted blanks in the request 
    # or database files.
    sed -e 's/  *//g' ${DATABASE_FILE} | sort | uniq > data_uniq_$$.txt
    sed -e 's/  *//g' cataloguefiles_$$.txt | sort |\
      uniq > catalogue_uniq_$$.txt
    cat data_uniq_$$.txt catalogue_uniq_$$.txt | sort | uniq -d >\
        dupl_images_$$.txt
    if [ -s dupl_images_$$.txt ]
    then
      # reject double lines by concatenating the
      # original list and the double entries and
      # printing only unique entries.
      cat catalogue_uniq_$$.txt dupl_images_$$.txt | sort | uniq -u >\
          reqfiles_$$.txt
    else
      cp catalogue_uniq_$$.txt reqfiles_$$.txt 
    fi 	
  else
    echo "${DATABASE_FILE} does not exist!"
    echo "no check for existing files possible!!"
    echo ""
    echo "All images will be retrieved/listed!"
    sort cataloguefiles_$$.txt | uniq > reqfiles_$$.txt 
  fi  
else
  echo "No database file given! All images will be retrieved/listed!"
  sort cataloguefiles_$$.txt | uniq > reqfiles_$$.txt 
fi

# Determine number of images to retrieve;
NREQIMAGES=`wc -l reqfiles_$$.txt | awk '{print $1}'`

# check if something needs to be requested/listed at all
if [ ${NREQIMAGES} -lt 1 ]
then
  echo "No images to request; it seems that you have all of them already!"
  cleanTmpFiles
  exit 0
fi

if [ "${LIST_ONLY}" = "N" ]
then
  if [ "${TRY_ANONYMOUS}" -eq 1 ]
  then
    WGET_PASSWD_OPT=""
  else
    WGET_PASSWD_OPT="--http-user=${CADC_USER} --http-passwd=${CADC_PASSWORD}"
  fi

  # handle output directory:
  if [ "${INPUT_IMAGE_LIST}" = "DUMMY" ]
  then
    OUTPUT_DIR=${OUTPUT_DIR}/${FIELD}/${FILTER}
  fi  

  if [ ! -d ${OUTPUT_DIR} ]
  then
    mkdir -p ${OUTPUT_DIR}\
      || { echo "Creation of output dir. ${OUTPUT_DIR} failed!!"; exit 1; }
  fi

  # delete 'old' files from the OUTPUT_DIR 
  # with the names of files to be requested.
  # Those are, most probably, files whose download
  # did not succeed and which we want to reretrieve.
  while read IMAGE
  do
    if [ -f ${OUTPUT_DIR}/${IMAGE}.fits.fz ]; then
      echo "Removing ${OUTPUT_DIR}/${IMAGE}.fits.fz"
      rm ${OUTPUT_DIR}/${IMAGE}.fits.fz
    fi
  done < reqfiles_$$.txt

  # now retrieve the images:
  echo "Retrieving ${NREQIMAGES} images with wget"
  ERRORSOCCUR=0
  while read IMAGE
  do
    wget --output-document=${OUTPUT_DIR}/${IMAGE}.fits.fz\
      ${WGET_PASSWD_OPT} \
      "http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/getData?archive=CFHT&file_id=${IMAGE}"
    if [ $? -ne 0 ]
    then
      ERRORSOCCUR=1
      # sometimes a faulty 'wget' call leaves behind an empty
      # result file; get rid of it:
      if [ -e ${OUTPUT_DIR}/${IMAGE}.fits.fz ] &&\
         [ ! -s ${OUTPUT_DIR}/${IMAGE}.fits.fz ]
      then
        rm ${OUTPUT_DIR}/${IMAGE}.fits.fz 
      fi
    fi
  done < reqfiles_$$.txt
       
  if [ ${ERRORSOCCUR} -ne 0 ]
  then
    echo "WARNING: wget call(s) to CADC reported errors"
  fi
 
  # check retrieved images with fitsverify for the number
  # extensions and for FITS convention errors. Append ok images
  # to the databse.
  while read IMAGE
  do
    VERIFYRESULT=`fitsverify ${OUTPUT_DIR}/${IMAGE}.fits.fz |\
    awk 'BEGIN {count = 0; nextension = 0;} 
         ($1 ~ /[0-9]+/ && \
           ($0 ~ /Primary Array/ ||\
            $0 ~ /Image Array/ ||\
            $0 ~ /Binary Table/)) {
           count_warnings = NF-1; 
           count_errors = NF; 
           count += $count_warnings + $count_errors; 
           if ($0 ~ /Image Array/ || $0 ~ /Binary Table/) {nextension++}
         } END {print nextension, count}'`

    NEXTEND=`echo ${VERIFYRESULT} | awk '{print $1}'`
    CURRERRORCODE=`echo ${VERIFYRESULT} | awk '{print $2}'`

    # The number of image extensions must be o.k.:
    if [ ${NEXTEND} -eq ${INST_EXTENSIONS} ]
    then
      ERROROK=`awk 'BEGIN {if ('"${ERROR_CODES}"' ~ /'${CURRERRORCODE}'/) { 
                       print 1; 
                     } else {
                       print 0;
                     }}'`  
      if [ ${ERROROK} -eq 0 ]
      then
        echo "${IMAGE}.fits.fz was not retrieved or has errors !!"	
      else
        echo "${IMAGE}.fits.fz is ok !!"
        if [ ${DATABASE_FILE} ]
        then
          echo "appending ${IMAGE} to database"
          echo ${IMAGE} >> ${DATABASE_FILE}
        fi
      fi
    else
      echo "${IMAGE}.fits.fz does not have ${INST_EXTENSIONS} extensions!"
    fi
  done < reqfiles_$$.txt
else
  echo "The following ${NREQIMAGES} files would be retrieved:"
  cat reqfiles_$$.txt
fi

# clean up and bye
cleanTmpFiles

exit 0
